/* Config.hh. Generated by CMake for ignition-gazebo. */

/* Version number */
#define IGNITION_GAZEBO_MAJOR_VERSION 6
#define IGNITION_GAZEBO_MINOR_VERSION 1
#define IGNITION_GAZEBO_PATCH_VERSION 0

#define IGNITION_GAZEBO_VERSION "6.1"
#define IGNITION_GAZEBO_VERSION_FULL "6.1.0"
#define IGNITION_GAZEBO_MAJOR_VERSION_STR "6"

#define IGNITION_GAZEBO_VERSION_NAMESPACE v6

#define IGNITION_GAZEBO_VERSION_HEADER "Ignition Gazebo, version 6.1.0\nCopyright (C) 2018 Open Source Robotics Foundation.\nReleased under the Apache 2.0 License.\n\n"

#define IGNITION_GAZEBO_GUI_CONFIG_PATH "D:/bld/libignition-gazebo6_1635251434674/_h_env/Library/share/ignition/ignition-gazebo6/gui"
#define IGNITION_GAZEBO_SYSTEM_CONFIG_PATH "D:/bld/libignition-gazebo6_1635251434674/_h_env/Library/share/ignition/ignition-gazebo6/systems"
#define IGNITION_GAZEBO_SERVER_CONFIG_PATH "D:/bld/libignition-gazebo6_1635251434674/_h_env/Library/share/ignition/ignition-gazebo6"
#define IGN_GAZEBO_PLUGIN_INSTALL_DIR "D:/bld/libignition-gazebo6_1635251434674/_h_env/Library/lib/ign-gazebo-6/plugins"
#define IGN_GAZEBO_GUI_PLUGIN_INSTALL_DIR "D:/bld/libignition-gazebo6_1635251434674/_h_env/Library/lib/ign-gazebo-6/plugins/gui"
#define IGN_GAZEBO_WORLD_INSTALL_DIR "D:/bld/libignition-gazebo6_1635251434674/_h_env/Library/share/ignition/ignition-gazebo6/worlds"

/* #undef IGNITION_GAZEBO_BUILD_TYPE_PROFILE */
/* #undef IGNITION_GAZEBO_BUILD_TYPE_DEBUG */
/* #undef IGNITION_GAZEBO_BUILD_TYPE_RELEASE */
