



if not exist %PREFIX%\\Library\\include\\ignition\\gazebo6\\ignition\\gazebo.hh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\\Library\\lib\\ignition-gazebo6.lib exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\\Library\\bin\\ignition-gazebo6.dll exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\\Library\\lib\\cmake\\ignition-gazebo6\\ignition-gazebo6-config.cmake exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
